"""
Cron parsing tests
"""

import unittest
from cronredux import cronparser


class ParserTests(unittest.TestCase):

    def test_parse_empty(self):
        cases = [
            (''),
            ('# * * * * * totally not running this\n'),
        ]
        for case in cases:
            with self.subTest(case):
                self.assertFalse(cronparser.parseline(case))

    def test_parse_simple_spec(self):
        valid = [
            ('* * * * *', 'command'),
            ('*/1 * * * *', 'command'),
            ('1,2,3 * * * *', 'command'),
            ('1,2,3 * * * *', ' command'),
            ('1,2,3 2,3 * * *', ' command '),
            ('1,2,3 * * * *', ' command more\n'),
        ]
        for x in valid:
            with self.subTest(x):
                self.assertEqual(cronparser.parseline('%s %s' % x), x)

    def test_parse_alias_spec(self):
        valid = [
            ('@daily', 'command'),
        ]
        for x in valid:
            with self.subTest(x):
                self.assertEqual(cronparser.parseline('%s %s' % x), x)

    def test_parse_string(self):
        s = '\n\n* * * * * command\n\n'
        self.assertEqual(list(cronparser.parses(s)),
                         [('* * * * *', 'command')])



"""
Diagnostic web server endpoints.
"""

import aiohttp_jinja2
import asyncio
import jinja2
import os
import pendulum
import platform
import shellish
from aiohttp import web


class DiagService(object):
    """ Diagnostic Web Service. """

    platform_info = {
        "system": platform.system(),
        "platform": platform.platform(),
        "node": platform.node(),
        "dist": ' '.join(platform.dist()),
        "python": platform.python_version()
    }
    ui_dir = os.path.join(os.path.dirname(__file__), 'ui')

    def __init__(self, tasks, args, loop, sched=None, plain=False):
        self.tasks = tasks
        self.args = args
        self.loop = loop
        self.tpl_context = {
            "environ": os.environ,
            "args": args,
            "tasks": tasks,
            "tasks_by_id": dict((x.ident, x) for x in tasks),
            "task_execs": lambda: dict(((x.task.ident, x.ident), x)
                                       for x in sched.history),
            "platform": self.platform_info,
            "ui_dir": self.ui_dir,
            "started": pendulum.now(),
            "pendulum": pendulum,
            "sched": sched
        }
        self.plain_output = plain

    @asyncio.coroutine
    def start(self):
        self.app = web.Application(loop=self.loop)
        tpl_loader = jinja2.FileSystemLoader(self.ui_dir)
        env = aiohttp_jinja2.setup(self.app, loader=tpl_loader)
        env.globals.update({
            "sorted": sorted
        })
        self.app.router.add_route('GET', '/', self.index_redir)
        self.app.router.add_route('GET', '/health', self.health)
        self.app.router.add_route('GET', '/ui/{path}', self.tpl_handler)
        self.app.router.add_static('/ui/static',
                                   os.path.join(self.ui_dir, 'static'))
        self.handler = self.app.make_handler()
        listen = self.args.diag_addr, self.args.diag_port
        self.server = yield from self.loop.create_server(self.handler, *listen)
        shellish.vtmlprint('<b>Running diag web server</b>: '
                           '<blue><u>http://%s:%s</u></blue>' % listen,
                           plain=self.plain_output)

    @asyncio.coroutine
    def index_redir(self, request):
        return web.HTTPFound('/ui/index.html')

    @asyncio.coroutine
    def health(self, request):
        return web.json_response({
            "platform_info": self.platform_info,
            "tasks": [x.cmd for x in self.tasks],
        })

    @asyncio.coroutine
    def tpl_handler(self, request):
        path = request.match_info['path']
        context = self.tpl_context.copy()
        context['request'] = request
        return aiohttp_jinja2.render_template(path, request, context)

    @asyncio.coroutine
    def cleanup(self):
        self.server.close()
        yield from self.server.wait_closed()
        yield from self.app.shutdown()
        yield from self.handler.finish_connections(1)
        yield from self.app.cleanup()



"""
Handle task monitoring and scheduling.
"""

import asyncio
import collections
import functools
import itertools
import pendulum
import shellish
import subprocess
import textwrap
import traceback


class TaskExecContext(object):
    """ A data structure representing a task execution.  Status and info about
    an invocation is kept here and these instances are used to track activity.
    """

    def __init__(self, task, loop):
        self.started = None
        self.finished = None
        self.returncode = None
        self.output = ''
        self.state = 'init'
        self.ident = next(task.context_identer)
        self.task = task
        self.loop = loop

    def __str__(self):
        return '<TaskExecContext %d [%s]> for %s' % (self.ident, self.state,
                                                     self.task)

    def set_start(self):
        assert self.state == 'init'
        self.started = pendulum.now()
        self.state = 'start'

    def set_finish(self, returncode, output):
        assert self.state == 'start'
        self.finished = pendulum.now()
        self.state = 'finish'
        self.output = output
        self.returncode = returncode

    @property
    def elapsed(self):
        """ Increments for active tasks and when finished gives the total
        execution time. """
        if self.started is None:
            return pendulum.Interval()
        if self.finished is None:
            return pendulum.now() - self.started
        else:
            return self.finished - self.started

    def is_error(self):
        if self.state != 'finish':
            raise TypeError('Task is still running')
        return not not self.returncode

    def is_done(self):
        return self.state == 'finished'

    @asyncio.coroutine
    def run_task(self):
        self.set_start()
        process, output = yield from self.task(self.loop)
        self.set_finish(process.returncode, output.decode())


class Task(object):
    """ Encapsulate a repeated task. """

    identer = itertools.count()

    def __init__(self, crontab, cmd):
        self.ident = next(self.identer)
        self.context_identer = itertools.count()
        self.crontab = crontab
        self.cmd = cmd
        self.last_eta = crontab.next(default_utc=False)
        self.run_count = 0
        self.elapsed = pendulum.Interval()

    def __str__(self):
        dots = shellish.beststr("…", '...')
        cmd = textwrap.shorten(self.cmd, width=40, placeholder=dots)
        return '<Task %d: cmd="%s">' % (self.ident, cmd)

    def next_run(self):
        """ Estimated time of next run. """
        return pendulum.Interval(seconds=self.crontab.next(default_utc=False))

    @asyncio.coroutine
    def __call__(self, loop):
        start = pendulum.now()
        ps = yield from asyncio.create_subprocess_shell(
            self.cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            loop=loop)
        output = (yield from ps.communicate())[0]
        self.elapsed += pendulum.now() - start
        self.run_count += 1
        return ps, output

    def ready(self):
        """ Returns true when this task should be run.  If this function is not
        called as or more frequently than the cron period then it will skip or
        even never return true.  It is advised to call this at least double the
        minimum cron period of 1 minute, e.g. 30 seconds. """
        next_eta = self.crontab.next(default_utc=False)
        r = self.last_eta < next_eta  # rollover occurred == ready
        self.last_eta = next_eta
        return r


class Scheduler(object):
    """ Manage execution and scheduling of tasks. """

    def __init__(self, tasks, args, notifier, loop):
        self.tasks = tasks
        self.args = args
        self.notifier = notifier
        self.loop = loop
        self.workers_sem = asyncio.Semaphore(self.args.max_concurrency)
        self.wakeup = asyncio.Event(loop=loop)
        self.active = []
        self.history = collections.deque(maxlen=100)

    def is_active(self, task):
        """ Scan task exec list looking for this task. """
        for x in self.active:
            if x.task is task:
                return True
        return False

    @asyncio.coroutine
    def run(self):
        """ Babysit the task scheduling process. """
        while True:
            for task in self.tasks:
                if task.ready():
                    if not self.args.allow_overlap and self.is_active(task):
                        yield from self.notifier.warning('Skipping `%s`' %
                                                         task,
                                                         'Previous task is '
                                                         'still active.')
                    else:
                        yield from self.workers_sem.acquire()
                        yield from self.enqueue_task(task)
            try:
                yield from asyncio.wait_for(self.wakeup.wait(), 1)
            except asyncio.TimeoutError:
                pass
            else:
                self.wakeup.clear()

    @asyncio.coroutine
    def enqueue_task(self, task):
        """ Create (and return) the task status and run the task in the
        background. """
        context = TaskExecContext(task, self.loop)
        self.active.append(context)
        self.history.appendleft(context)
        f = self.loop.create_task(self.task_runner(context))
        f.add_done_callback(functools.partial(self.on_task_done, context))

    @asyncio.coroutine
    def task_runner(self, context):
        """ Background runner for actually executing the command. """
        if self.args.notify_exec:
            yield from self.notifier.info('Starting: `%s`' % context.task,
                                          footer='Exec #%d' % context.ident)
        yield from context.run_task()
        footer = 'Exec #%d - Duration %s' % (context.ident,
                                             context.elapsed)
        if context.returncode:
            yield from self.notifier.error('Failed: `%s`' % context.task,
                                           raw=context.output, footer=footer)
        elif self.args.notify_exec:
            yield from self.notifier.info('Succeeded: `%s`' % context.task,
                                          raw=context.output, footer=footer)
        for x in context.output.splitlines():
            print('[%s] [job:%d] [exit:%d] %s' % (context.task.cmd,
                  context.ident, context.returncode, x))

    def on_task_done(self, context, f):
        try:
            f.result()
        except Exception as e:
            traceback.print_exc()
            raise SystemExit("Unrecoverable Error: %s" % e)
        finally:
            self.active.remove(context)
            self.workers_sem.release()
            self.wakeup.set()



"""
A newage cron service.
"""

import asyncio
import crontab
import shellish
import cronredux
from cronredux import cronparser, notification, scheduler
from cronredux.diag import web


class CronReduxCommand(shellish.Command):
    """ Cronredux.

    A reimagined cron executor. """

    name = "cronredux"

    def setup_args(self, parser):
        self.add_file_argument('crontab', help='Crontab file to read.')
        self.add_argument('--diag-addr', default='0.0.0.0', help='Port for '
                          'diagnostic web server.')
        self.add_argument('--diag-port', type=int, default=7907, help='Port '
                          'for diagnostic web server.')
        self.add_argument('--verbose', action='store_true')
        self.add_argument('--slack-webhook', help='WebHook URL for slack '
                          'notifications.')
        self.add_argument('--slack-channel', help='Override the default '
                          'slack channel.')
        self.add_argument('--slack-username', help='Override the default '
                          'slack username.')
        self.add_argument('--slack-icon-emoji', help='Override the default '
                          'slack icon emoji.')
        self.add_argument('--notify-exec', action='store_true',
                          help='Enable notifications for each task execution.')
        self.add_argument('--XXX-slow-exec-warning', type=float, default=60,
                          help='Time in seconds before a warning is generated '
                          'about slow task execution.')
        self.add_argument('--XXX-backlog-warning', type=float, default=60,
                          help='Time in seconds before a warning is generated '
                          'about a delayed task execution due to a backlog.')
        self.add_argument('--max-concurrency', type=int, default=10,
                          help='Max tasks that will be allowed to'
                          'run concurrently.')
        self.add_argument('--allow-overlap', action='store_true')
        self.add_argument('--plain', action='store_true')

    def run(self, args):
        self.args = args
        tasks = []
        if args.plain:
            cronredux.PLAIN_OUTPUT = True
        if args.verbose:
            shellish.vtmlprint("<b>Processing crontab file:</b> <red>%s</red>"
                               % args.crontab, plain=cronredux.PLAIN_OUTPUT)
        with args.crontab as f:
            for spec, command in cronparser.parsef(f):
                if args.verbose:
                    shellish.vtmlprint("<b>Adding task:</b> <blue>%s</blue> %s"
                                       % (spec, command),
                                       plain=cronredux.PLAIN_OUTPUT)
                tasks.append(scheduler.Task(crontab.CronTab(spec), command))
        if args.slack_webhook:
            notifier = notification.SlackNotifier(
                args.slack_webhook,
                channel=args.slack_channel,
                username=args.slack_username,
                icon_emoji=args.slack_icon_emoji)
        else:
            notifier = notification.PrintNotifier()
        loop = asyncio.get_event_loop()
        sched = scheduler.Scheduler(tasks, args, notifier, loop)
        diag = web.DiagService(tasks,
                               args,
                               loop,
                               sched=sched,
                               plain=cronredux.PLAIN_OUTPUT)
        try:
            loop.run_until_complete(notifier.setup(loop))
            loop.run_until_complete(diag.start())
            loop.create_task(sched.run())
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            if args.verbose:
                shellish.vtmlprint("<b>Shutting Down</b>",
                                   plain=cronredux.PLAIN_OUTPUT)
                loop.run_until_complete(diag.cleanup())
            loop.close()


def main():
    CronReduxCommand()()


if __name__ == '__main__':
    main()



"""
Read and parse crontab files.
"""


def parsef(f):
    """ Parse a file object into a generator of (spec, command) tuples. """
    for x in f:
        parsed = parseline(x.rstrip('\n'))
        if parsed is not None:
            yield parsed


def parses(s):
    """ Parse a multiline string object into a generator of (spec, command)
    tuples. """
    for x in s.splitlines():
        parsed = parseline(x)
        if parsed is not None:
            yield parsed


def parseline(line):
    """ Extract the spec and command from a line.  If the line is any kind
    of no-op we return None. """
    if not line:
        return
    # return None if line is commented
    if line[0] == '#':
        return
    if line[0] == '@':
        return tuple(line.split(' ', 1))
    else:
        parts = line.split(' ', 5)
        return ' '.join(parts[:5]), parts[-1]



"""
Support for notifications.
"""

import aiohttp
import asyncio
import json
import shellish
import time
import cronredux


class PrintNotifier(object):
    """ A basic interface for sending notifications to stdout. """

    @asyncio.coroutine
    def setup(self, loop):
        pass

    @asyncio.coroutine
    def info(self, title, message='', raw='', footer=None):
        shellish.vtmlprint("<b><blue>INFO: %s</blue></b>" % title,
                           plain=cronredux.PLAIN_OUTPUT)
        if message or raw:
            for line in (message + raw).splitlines():
                shellish.vtmlprint("    <blue>%s</blue>" % line,
                                   plain=cronredux.PLAIN_OUTPUT)
        if footer:
            shellish.vtmlprint("    <dim>%s</dim>" % footer,
                               plain=cronredux.PLAIN_OUTPUT)

    @asyncio.coroutine
    def warning(self, title, message='', raw='', footer=None):
        shellish.vtmlprint("<b>WARN: %s</b>" % title,
                           plain=cronredux.PLAIN_OUTPUT)
        if message or raw:
            for line in (message + raw).splitlines():
                shellish.vtmlprint("    %s" % line,
                                   plain=cronredux.PLAIN_OUTPUT)
        if footer:
            shellish.vtmlprint("<dim>%s</dim>" % footer,
                               plain=cronredux.PLAIN_OUTPUT)

    @asyncio.coroutine
    def error(self, title, message='', raw='', footer=None):
        shellish.vtmlprint("<b><red>ERROR: %s</red></b>" % title,
                           plain=cronredux.PLAIN_OUTPUT)
        if message or raw:
            for line in (message + raw).splitlines():
                shellish.vtmlprint("    <red>%s</red>" % line,
                                   plain=cronredux.PLAIN_OUTPUT)
        if footer:
            shellish.vtmlprint("    <dim>%s</dim>" % footer,
                               plain=cronredux.PLAIN_OUTPUT)


class SlackNotifier(object):
    """ Send messages to a slack webhook. """

    default_username = 'Cronredux'
    default_icon_emoji = ':calendar:'
    max_raw_size = 700

    def __init__(self, webhook_url, channel=None, username=None,
                 icon_emoji=None):
        self.webhook = webhook_url
        self.default_payload = p = {}
        if channel is not None:
            p['channel'] = channel
        p['username'] = username if username is not None else \
                        self.default_username
        p['icon_emoji'] = icon_emoji if icon_emoji is not None else \
                        self.default_icon_emoji

    @asyncio.coroutine
    def setup(self, loop):
        self.loop = loop
        headers = {'content-type': 'application/json'}
        self.session = aiohttp.ClientSession(loop=loop, headers=headers)
        self.lock = asyncio.Lock()

    @asyncio.coroutine
    def post(self, data):
        payload = self.default_payload.copy()
        payload.update(data)
        with (yield from self.lock):
            while True:
                resp = yield from self.session.post(self.webhook,
                                                    data=json.dumps(payload))
                if resp.status != 200:
                    content = yield from resp.json()
                    if resp.status == 429:  # rate limited
                        retry = content['retry_after']
                        shellish.vtmlprint('<red>Slack Rate Limit Reached: '
                                           'Retry in %f seconds.</red>' % retry,
                                           plain=cronredux.PLAIN_OUTPUT)
                        yield from asyncio.sleep(retry)
                        continue
                    else:
                        shellish.vtmlprint('<b><red>Slack Post Failed:</red> '
                                           '(%d) - %s</b>' % (resp.status,
                                           content),
                                           plain=cronredux.PLAIN_OUTPUT)
                yield from resp.release()
                break

    @asyncio.coroutine
    def log(self, level, color, title, message=None, raw=None, footer=None):
        text = []
        if message:
            text.append(message)
        if raw:
            if len(raw) > self.max_raw_size:
                head = raw[:self.max_raw_size // 2]
                tail = raw[-self.max_raw_size // 2:]
                text.append('```%s```\n>..._contents clipped_...\n```%s```' % (
                            head, tail))
            else:
                text.append('```%s```' % raw)
        payload = {
            "color": color,
            "pretext": '*%s*' % title,
            "fallback": title,
            "ts": time.time(),
            "mrkdwn_in": ['text', 'pretext', 'footer']
        }
        if text:
            payload['text'] = '\n'.join(text)
        if footer:
            payload['footer'] = footer
        yield from self.post({"attachments": [payload]})

    @asyncio.coroutine
    def info(self, *args, **kwargs):
        yield from self.log('INFO', '#2200cc', *args, **kwargs)

    @asyncio.coroutine
    def warning(self, *args, **kwargs):
        yield from self.log('WARN', '#ffcc44', *args, **kwargs)

    @asyncio.coroutine
    def error(self, *args, **kwargs):
        yield from self.log('ERROR', '#cc4444', *args, **kwargs)



"""
Test scheduler logic.
"""

import time
from cronredux import scheduler


class TestTask(scheduler.Task):

    def __init__(self, *args, on_call=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.state_hist = []
        self.state_ts_hist = []
        self.set_state('init')
        self.on_call = on_call
        self.run_count = 0

    def set_state(self, value):
        self.state_hist.append(value)
        self.state_ts_hist.append(time.monotonic())

    def __call__(self):
        self.set_state('run')
        if self.on_call is not None:
            self.on_call()
        self.set_state('done')



import os
from setuptools import setup, find_packages

README = 'README.md'

cronredux_version = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), 'cronredux', '__init__.py')

exec(compile(open(cronredux_version, 'r').read(), cronredux_version, 'exec'))

with open('requirements.txt') as f:
    requirements = f.readlines()

long_description = None
with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='cronredux',
    version=VERSION,
    description='A reimagined cron executor',
    author='HomeCU',
    author_email='developers@homecu.com',
    url='https://github.com/homecu/cronredux/',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    test_suite='test',
    install_requires=requirements,
    entry_points={
        'console_scripts': ['cronredux=cronredux.main:main'],
    },
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)
